from pyrfuniverse.envs import RFUniverseBaseEnv
import numpy as np


class FrankaGraspEnv(RFUniverseBaseEnv):

    def __init__(self, executable_file=None):
        super().__init__(
            executable_file,
            camera_channel=True,
            rigidbody_channel=True,
            articulation_channel=True
        )
        self.camera_channel.set_action(
            'AddCamera',
            position=[0, 0, 0],
            rotation=[0, 0, 0],
            parent_camera_idx=0
        )
        self._step()

        self.camera_channel.set_action(
            'AddCamera',
            position=[0, 0, 0],
            rotation=[0, 0, 0],
            parent_camera_idx=0
        )
        self._step()

        self.camera_channel.set_action(
            'ResetCamera',
            index=2,
            position=[0, 0.1, 0]
        )
        self._step()

        # self.asset_channel.set_action(
        #     'LoadRigidbody',
        #     filename='D:\\GitHub\\rfuniverse\\RFUniverse\\Assets\\AssetBundles\\rigidbody',
        #     name='011_banana',
        #     position=[-0.612053275,0.628000021,-0.356999993]
        # )
        # self._step()

        self.robot_joint_positions = np.array([0.0 for i in range(8)], dtype=float)

    def step(self, a: np.ndarray):
        """
        Params:
            a: 8-d or 16-d numpy array. The first 7 dimensions are for each joint's position of Franka (-180~180),
               while the 8th dimension is for gripper's width. If the total dimension is 16, the last
               8 dimensions are the speed scales. Otherwise, the speed scale will be default to 1.0
        """
        assert a.shape == (16,) or a.shape == (8,), \
            'The shape of action must be (14,) or (7,), but got {}'.format(a.shape)
        if a.shape ==  (8,):
            a = np.concatenate((a, np.array([1.0 for i in range(8)], dtype=float)))

        self._set_franka_joints(a)
        self._wait_for_moving()
        self.t += 1

        return self._get_obs()

    def _get_obs(self):
        # Read message from channels
        return self.rigidbody_channel.data

    def reset(self):
        self.t = 0
        self.r = 0
        self.env.reset()

        return self._get_obs()

    def seed(self):
        pass

    def _set_franka_joints(self, a: np.ndarray):
        self.articulation_channel.set_action(
            'SetJointPosition',
            index=0,
            joint_positions=list(a[0:7]),
            speed_scales=list(a[8:15])
        )
        self._step()

        # Since a[7] is gripper's width, convert it to joint position value
        a[7] = -1 * a[7] / 2

        self.articulation_channel.set_action(
            'SetJointPosition',
            index=1,
            joint_positions=[a[7], a[7]],
            speed_scales=[a[15], a[15]]
        )
        self._step()

    def _control_gripper(self, joint_positions: np.ndarray, speed: np.ndarray):
        self.articulation_channel.set_action(
            'SetJointPosition',
            index=1,
            joint_positions=list(joint_positions),
            speed_scales=list(speed)
        )

    def _open_gripper(self, speed: float=1.0):
        self._control_gripper(
            joint_positions=np.array([-0.04, -0.04], dtype=float),
            speed=np.array([speed, speed], dtype=float)
        )

    def _close_gripper(self, speed: float=1.0):
        self._control_gripper(
            joint_positions=np.array([0, 0], dtype=float),
            speed=np.array([speed, speed], dtype=float)
        )

    def _wait_for_moving(self):
        self.articulation_channel.data.clear()
        while not 0 in self.articulation_channel.data.keys():
            self._step()
        while not (self.articulation_channel.data[0]['all_stable'] and self.articulation_channel.data[1]['all_stable']):
            self._step()



# Version of the library that will be used to upload to pypi
__version__ = "0.9.10"

import os.path
import json

def read_config():
    if not os.path.exists(config_path):
        config = {}
        config['assets_path'] = ''
        config['executable_file'] = ''
        with open(config_path, 'w', encoding='utf-8') as file:
            json.dump(config, file, indent=True)
    with open(config_path, 'r', encoding='utf-8') as file:
        return json.load(file)


user_path = os.path.expanduser('~/.rfuniverse')
if not os.path.exists(user_path):
    os.makedirs(user_path)
config_path = os.path.join(user_path, 'config.json')
config = read_config()
assets_path = config['assets_path']
executable_file = config['executable_file']



import struct
import socket
import threading
from sys import platform

import numpy as np


class RFUniverseCommunicator(threading.Thread):
    def __init__(self, port: int = 5004, is_async: bool = False, receive_data_callback=None):
        threading.Thread.__init__(self)
        self.is_async = is_async
        # self.sync_send_bytes_queue = []
        self.read_offset = 0
        self.on_receive_data = receive_data_callback

        self.server = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.server.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        self.server.bind(('localhost', port))
        print(f'Waiting for connections on port: {port}...')
        self.server.listen(1)
        self.client, self.addr = self.server.accept()
        print(f'Connected successfully')
        self.client.settimeout(None)
        self.client.setsockopt(socket.SOL_SOCKET, socket.SO_SNDBUF, 1024 * 1024 * 10)
        self.client.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, 1024 * 1024 * 10)
        self.client.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
        if self.is_async:
            self.start()
        else:
            self.receive_step()

    def run(self):
        while True:
            data = self.receive_bytes()
            objs = self.receive_object(data)
            if self.on_receive_data is not None:
                self.on_receive_data(objs)

    def sync_step(self):
        if self.is_async:
            return
        self.send_object("StepStart")
        # for item in self.sync_send_bytes_queue:
            # self.send_bytes(item)
        # self.sync_send_bytes_queue.clear()
        self.receive_step()

    def receive_step(self):
        # sync_receive_objects_queue = []
        while True:
            data = self.receive_bytes()
            objs = self.receive_object(data)
            if len(objs) > 0 and objs[0] == "StepEnd":
                break
            self.on_receive_data(objs)
        #     sync_receive_objects_queue.append(objs)
        # if self.on_receive_data is not None:
        #     for item in sync_receive_objects_queue:
        #         self.on_receive_data(item)
        # sync_receive_objects_queue.clear()

    def receive_bytes(self):
        data = self.client.recv(4)
        length = int.from_bytes(data, byteorder='little', signed=False)
        return self.client.recv(length)

    def send_bytes(self, data: bytes):
        length = len(data).to_bytes(4, byteorder='little', signed=False)
        self.client.send(length)
        self.client.send(data)
        if platform == 'linux':
            self.client.setsockopt(socket.IPPROTO_TCP, socket.TCP_QUICKACK, 1)

    def receive_object(self, data: bytes) -> list:
        self.read_offset = 0
        count = self.read_int(data)
        objs = []
        for i in range(count):
            objs.append(self.read_object(data))
        return objs

    def read_object(self, datas: bytes) -> object:
        data_type = self.read_string(datas)
        if data_type == 'int':
            return self.read_int(datas)
        elif data_type == 'float':
            return self.read_float(datas)
        elif data_type == 'string':
            return self.read_string(datas)
        elif data_type == 'bool':
            return self.read_bool(datas)
        elif data_type == 'bytes':
            return self.read_bytes(datas)
        elif data_type == 'vector3':
            return self.read_object(datas)
        elif data_type == 'quaternion':
            return self.read_object(datas)
        elif data_type == 'matrix':
            return self.read_object(datas)
        elif data_type == 'rect':
            return [self.read_float(datas) for _ in range(4)]
        elif data_type == 'array':
            rank = self.read_int(datas)
            shape = []
            for _ in range(rank):
                shape.append(self.read_int(datas))
            result = np.ndarray(shape, dtype=np.float32)
            result = result.reshape(-1)
            for i in range(len(result)):
                result[i] = self.read_float(datas)
            return result.reshape(shape)
        elif data_type == 'list':
            count = self.read_int(datas)
            result = []
            for _ in range(count):
                result.append(self.read_object(datas))
            return result
        elif data_type == 'dict':
            count = self.read_int(datas)
            result = {}
            for _ in range(count):
                key = self.read_object(datas)
                value = self.read_object(datas)
                result[key] = value
            return result
        elif data_type == 'tuple':
            count = self.read_int(datas)
            result = []
            for _ in range(count):
                result.append(self.read_object(datas))
            return tuple(result)
        elif data_type == 'null' or data_type == 'none':
            return None
        else:
            print(f'dont support this type: {data_type}')
            return None

    def read_string(self, datas: bytes) -> str:
        count = self.read_int(datas)
        self.read_offset += count
        return datas[self.read_offset - count:self.read_offset].decode('utf-8')

    def read_int(self, datas: bytes) -> int:
        self.read_offset += 4
        return int.from_bytes(datas[self.read_offset - 4:self.read_offset], byteorder='little')

    def read_float(self, datas: bytes) -> float:
        self.read_offset += 4
        return struct.unpack('f', datas[self.read_offset - 4:self.read_offset])[0]

    def read_bool(self, datas: bytes) -> bool:
        self.read_offset += 1
        return bool(int.from_bytes(datas[self.read_offset - 1:self.read_offset], byteorder='little'))

    def read_bytes(self, datas: bytes) -> bytes:
        count = self.read_int(datas)
        self.read_offset += count
        return datas[self.read_offset - count:self.read_offset]

    def send_object(self, *args):
        datas = bytearray()
        self.write_int(datas, len(args))
        for obj in args:
            self.write_object(datas, obj)

        # if self.is_async:
        self.send_bytes(bytes(datas))
        # else:
            # self.sync_send_bytes_queue.append(bytes(datas))

    def write_object(self, datas: bytearray, obj):
        if obj is None:
            self.write_string(datas, 'none')
        elif type(obj) == int or type(obj) == np.int32 or type(obj) == np.int64:
            self.write_string(datas, 'int')
            self.write_int(datas, obj)
        elif type(obj) == float or type(obj) == np.float32 or type(obj) == np.float64:
            self.write_string(datas, 'float')
            self.write_float(datas, obj)
        elif type(obj) == bool:
            self.write_string(datas, 'bool')
            self.write_bool(datas, obj)
        elif type(obj) == str:
            self.write_string(datas, 'string')
            self.write_string(datas, obj)
        elif type(obj) == bytes or type(obj) == bytearray:
            self.write_string(datas, 'bytes')
            self.write_bytes(datas, bytes(obj))
        elif type(obj) == list:
            self.write_string(datas, 'list')
            self.write_int(datas, len(obj))
            for item in obj:
                self.write_object(datas, item)
        elif type(obj) == dict:
            self.write_string(datas, 'dict')
            self.write_int(datas, len(obj))
            for item in obj:
                self.write_object(datas, item)
                self.write_object(datas, obj[item])
        elif type(obj) == np.ndarray:
            self.write_string(datas, 'array')
            self.write_int(datas, len(obj.shape))
            for i in range(len(obj.shape)):
                self.write_int(datas, obj.shape[i])
            obj = obj.reshape(-1)
            for i in range(len(obj)):
                self.write_float(datas, float(obj[i]))
        elif type(obj) == tuple:
            self.write_string(datas, 'tuple')
            self.write_int(datas, len(obj))
            for i in range(len(obj)):
                self.write_object(datas, obj[i])
        else:
            print(f'dont support this type: {type(obj)}')
            self.write_string(datas, 'null')

    def write_string(self, datas: bytearray, s: str):
        s_byte = s.encode('utf-8')
        self.write_int(datas, len(s_byte))
        datas.extend(s_byte)

    def write_int(self, datas: bytearray, i: int):
        datas.extend(i.to_bytes(4, byteorder='little'))

    def write_float(self, datas: bytearray, f: float):
        datas.extend(struct.pack('f', f))

    def write_bool(self, datas: bytearray, b: bool):
        datas.extend(int(b).to_bytes(1, byteorder='little'))

    def write_bytes(self, datas: bytearray, b: bytes):
        self.write_int(datas, len(b))
        datas.extend(b)



from pyrfuniverse.envs import RFUniverseBaseEnv


class RfuniverseRobotHubEnv(RFUniverseBaseEnv):
    def __init__(self, executable_file=None):
        super().__init__(
            executable_file=executable_file,
            camera_channel=True,
            articulation_channel=True,
            game_object_channel=True,
        )

    def step(self):
        # In each time step, this function must be called to make sure Unity works well.
        self._step()

    def reset(self):
        self.env.reset()



import numpy as np
import pyrfuniverse.attributes as attr
import base64


class CameraAttr(attr.BaseAttr):
    """
    Camera attribute class, which can capture many kinds of screenshot
    of the scene in RFUniverse.
    """
    def parse_message(self, data: dict):
        """
        Parse messages. This function is called by internal function.

        Returns:
            Dict: A dict containing useful information of this class.
        
            self.data['width']: The width of image.

            self.data['height']: The height of image.

            self.data['fov']: The field of view of camera.

            self.data['rgb']: The bytes of rgb image.

            self.data['normal']: The bytes of normal image.

            self.data['id_map']: The bytes of instance segmentation mask image.

            self.data['depth']: The bytes of depth image.

            self.data['depth_exr']: The bytes of depth image in exr format.

            self.data['amodal_mask']: The bytes of amodal mask image.

            self.data['heat_map']: The bytes of heat map image.

            self.data['2d_bounding_box']: The 2d bouding box of objects in camera (image) coordinate.

            self.data['3d_bounding_box']: The 3d bounding box of objects in world coordinate.
        """
        super().parse_message(data)
        if 'rgb' in self.data:
            self.data['rgb'] = base64.b64decode(self.data['rgb'])
        if 'normal' in self.data:
            self.data['normal'] = base64.b64decode(self.data['normal'])
        if 'id_map' in self.data:
            self.data['id_map'] = base64.b64decode(self.data['id_map'])
        if 'depth' in self.data:
            self.data['depth'] = base64.b64decode(self.data['depth'])
        if 'depth_exr' in self.data:
            self.data['depth_exr'] = base64.b64decode(self.data['depth_exr'])
        if 'amodal_mask' in self.data:
            self.data['amodal_mask'] = base64.b64decode(self.data['amodal_mask'])
        if 'heat_map' in self.data:
            self.data['heat_map'] = base64.b64decode(self.data['heat_map'])

    def AlignView(self):
        """
        Make the camera in RFUniverse align the current view in GUI.
        """
        self._send_data('AlignView')

    def GetRGB(self, width: int = 512, height: int = 512, fov: float = 60., intrinsic_matrix: np.ndarray = None):
        """
        Get the camera RGB image.

        Args:
            width: Int, the width of image.
            height: Int, the height of image.
            fov: Float, the field of view for camera.
            intrinsic_matrix: A ndarray of shape 3*3, representing the camera intrinsic matrix. When this parameter is passed, `width`, `height` and `fov` will be ignroed.
        """
        self._send_data('GetRGB', intrinsic_matrix, int(width), int(height), float(fov))

    def GetNormal(self, width: int = 512, height: int = 512, fov: float = 60., intrinsic_matrix: np.ndarray = None):
        """
        Get the normal image in world coordinate.

        Args:
            width: Int, the width of image.
            height: Int, the height of image.
            fov: Float, the field of view for camera.
            intrinsic_matrix: A ndarray of shape 3*3, representing the camera intrinsic matrix. When this parameter is passed, `width`, `height` and `fov` will be ignroed.
        """
        self._send_data('GetNormal', intrinsic_matrix, int(width), int(height), float(fov))

    def GetID(self, width: int = 512, height: int = 512, fov: float = 60., intrinsic_matrix: np.ndarray = None):
        """
        Get the instance segmentation mask image. The color for each pixel is computed from object ID, see `pyrfuniverse.utils.rfuniverse_utility.GetColorFromID` for more details.

        Args:
            width: Int, the width of image.
            height: Int, the height of image.
            fov: Float, the field of view for camera.
            intrinsic_matrix: A ndarray of shape 3*3, representing the camera intrinsic matrix. When this parameter is passed, `width`, `height` and `fov` will be ignroed.
        """
        self._send_data('GetID', intrinsic_matrix, int(width), int(height), float(fov))

    def GetDepth(self, zero_dis: float, one_dis: float, width: int = 512, height: int = 512, fov: float = 60., intrinsic_matrix: np.ndarray = None):
        """
        Get the depth image from camera. Since eacg pixel of depth image returned from this function is 8-bit, user should limit the depth range (`zero_dis` and `one_dis`) for more accurate results.

        Args:
            zero_dis: The minimum distance in calculation.
            one_dis: The maximum distance in calculation.
            width: Int, the width of image.
            height: Int, the height of image.
            fov: Float, the field of view for camera.
            intrinsic_matrix: A ndarray of shape 3*3, representing the camera intrinsic matrix. When this parameter is passed, `width`, `height` and `fov` will be ignroed.
        """
        self._send_data('GetDepth', float(zero_dis), float(one_dis), intrinsic_matrix, int(width), int(height), float(fov))

    def GetDepthEXR(self, width: int = 512, height: int = 512, fov: float = 60., intrinsic_matrix: np.ndarray = None):
        """
        Get the depth image from camera. This function returns EXR format image bytes and each pixel is 32-bit.

        Args:
            width: Int, the width of image.
            height: Int, the height of image.
            fov: Float, the field of view for camera.
            intrinsic_matrix: A ndarray of shape 3*3, representing the camera intrinsic matrix. When this parameter is passed, `width`, `height` and `fov` will be ignroed.
        """
        self._send_data('GetDepthEXR', intrinsic_matrix, int(width), int(height), float(fov))

    def GetAmodalMask(self, target_id: int, width: int = 512, height: int = 512, fov: float = 60, intrinsic_matrix: np.ndarray = None):
        """
        Get the amodal mask image for target object.

        Args:
            target_id: The target object ID.
            width: Int, the width of image.
            height: Int, the height of image.
            fov: Float, the field of view for camera.
            intrinsic_matrix: A ndarray of shape 3*3, representing the camera intrinsic matrix. When this parameter is passed, `width`, `height` and `fov` will be ignroed.
        """
        self._send_data('GetAmodalMask', int(target_id), intrinsic_matrix, int(width), int(height), float(fov))

    def StartHeatMapRecord(self, targets_id: list):

        targets_id = [int(i) for i in targets_id]

        self._send_data('StartHeatMapRecord', targets_id)

    def EndHeatMapRecord(self):
        self._send_data('EndHeatMapRecord')

    def GetHeatMap(self, width: int = 512, height: int = 512, radius: int = 50, fov: float = 60., intrinsic_matrix: np.ndarray = None):
        """
        Get the heat map image.

        Args:
            width: Int, the width of image.
            height: Int, the height of image.
            radius: The radius of heat map.
            fov: Float, the field of view for camera.
            intrinsic_matrix: A ndarray of shape 3*3, representing the camera intrinsic matrix. When this parameter is passed, `width`, `height` and `fov` will be ignroed.
        """
        self._send_data('GetHeatMap', int(radius), intrinsic_matrix, int(width), int(height), float(fov))

    def Get2DBBox(self, width: int = 512, height: int = 512, fov: float = 60., intrinsic_matrix: np.ndarray = None):
        """
        Get the 2d bounding box of objects in current camera view.

        Args:
            width: Int, the width of image.
            height: Int, the height of image.
            radius: The radius of heat map.
            fov: Float, the field of view for camera.
            intrinsic_matrix: A ndarray of shape 3*3, representing the camera intrinsic matrix. When this parameter is passed, `width`, `height` and `fov` will be ignroed.
        """
        self._send_data('Get2DBBox', intrinsic_matrix, int(width), int(height), float(fov))

    def Get3DBBox(self):
        """
        Get the 3d bounding box of objects in world coordinate.
        """
        self._send_data('Get3DBBox')




# from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
# try:
#     from pyrfuniverse.envs.gym_wrapper_env import RFUniverseGymWrapper
#     from pyrfuniverse.envs.gym_goal_wrapper import RFUniverseGymGoalWrapper
# except ImportError:
#     pass
# from pyrfuniverse.envs.franka_grasp_env import FrankaGraspEnv
# from pyrfuniverse.envs.franka_push_env import FrankaPushEnv
# from pyrfuniverse.envs.balance_ball_env import BalanceBallEnv
# from pyrfuniverse.envs.balance_ball_env import BalanceBallEnvV0
# from pyrfuniverse.envs.bouncer_env import BouncerEnv
# from pyrfuniverse.envs.bouncer_env import BouncerEnvV0
# from pyrfuniverse.envs.roller_env import RollerEnv
# from pyrfuniverse.envs.roller_env import RollerEnvV0
# from pyrfuniverse.envs.gripper_nail import NailCardEnv
# from pyrfuniverse.envs.gripper_nail import NailCanEnv
# from pyrfuniverse.envs.multi_agent_navigation_env import MultiAgentNavigationEnv
# from pyrfuniverse.envs.tobor_robotiq85_manipulation_env import ToborRobotiq85ManipulationEnv
# from pyrfuniverse.envs.ur5_box_env import Ur5BoxEnv
# from pyrfuniverse.envs.ur5_drawer_env import Ur5DrawerEnv

# Other projects based on RFUniverse
# import pyrfuniverse.envs.robotube
# import pyrfuniverse.envs.rcareworld

__all__ = [
    'RFUniverseBaseEnv', 'FrankaGraspEnv', 'FrankaPushEnv', 'BalanceBallEnv',
    'RFUniverseGymWrapper', 'BalanceBallEnvV0', 'RFUniverseGymGoalWrapper',
    'BouncerEnv', 'BouncerEnvV0', 'RollerEnv', 'RollerEnvV0', 'NailCardEnv',
    'MultiAgentNavigationEnv', 'ToborRobotiq85ManipulationEnv', 'Ur5BoxEnv',
    'Ur5DrawerEnv',
]


rfuniverse_build_base_root = '/home/haoyuan/workspace/rfuniverse/rfuniverse/rfuniverse_build/'

try:
    from gym.envs.registration import register

    register(
        id='BalanceBallEnv-v0',
        entry_point='pyrfuniverse.envs:BalanceBallEnvV0'
    )

    register(
        id='BouncerEnv-v0',
        entry_point='pyrfuniverse.envs:BouncerEnvV0'
    )

    register(
        id='RollerEnv-v0',
        entry_point='pyrfuniverse.envs:RollerEnvV0'
    )

    register(
        id='MultiAgentNavigation-v1',
        entry_point='pyrfuniverse.envs:MultiAgentNavigationEnv',
        kwargs={
            'executable_file': rfuniverse_build_base_root + 'MultiAgentNavigationServer/RFUniverse.x86_64',
            'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/rigidbody',
            'num_agents': 5,
        },
    )

    register(
        id='MultiAgentNavigation-v2',
        entry_point='pyrfuniverse.envs:MultiAgentNavigationEnv',
        kwargs={
            'executable_file': rfuniverse_build_base_root + 'MultiAgentNavigationServer/RFUniverse.x86_64',
            'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/rigidbody',
            'num_agents': 5,
            'reset_on_collision': True
        },
    )

    register(
        id='MultiAgentNavigation-v3',
        entry_point='pyrfuniverse.envs:MultiAgentNavigationEnv',
        kwargs={
            'executable_file': rfuniverse_build_base_root + 'MultiAgentNavigationServer/RFUniverse.x86_64',
            'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/rigidbody',
            'num_agents': 3,
        },
    )

    register(
        id='MultiAgentNavigation-v4',
        entry_point='pyrfuniverse.envs:MultiAgentNavigationEnv',
        kwargs={
            'executable_file': rfuniverse_build_base_root + 'MultiAgentNavigationServer/RFUniverse.x86_64',
            'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/rigidbody',
            'num_agents': 3,
            'reset_on_collision': True
        },
    )

    # Franka robotics
    for reward_type in ['sparse', 'dense']:
        suffix = 'Dense' if reward_type == 'dense' else ''
        kwargs = {
            'executable_file': rfuniverse_build_base_root + 'FrankaRoboticsServer/RFUniverse.x86_64',
            'max_episode_length': 50,
            'reward_type': reward_type,
            'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/rigidbody'
        }

        register(
            id='FrankaPickAndPlace{}-v1'.format(suffix),
            entry_point='pyrfuniverse.envs.robotics:FrankaPickAndPlaceEnv',
            kwargs=kwargs,
            max_episode_steps=50,
        )

        register(
            id='FrankaReach{}-v1'.format(suffix),
            entry_point='pyrfuniverse.envs.robotics:FrankaReachEnv',
            kwargs=kwargs,
            max_episode_steps=50,
        )

        register(
            id='FrankaPush{}-v1'.format(suffix),
            entry_point='pyrfuniverse.envs.robotics:FrankaPushEnv',
            kwargs=kwargs,
            max_episode_steps=50,
        )

    # Franka-Cloth env
    franka_cloth_kwargs = {
        'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/obi_cloth',
        'executable_file': rfuniverse_build_base_root + 'FrankaCloth/RFUniverse.x86_64',
        'reward_type': 'dense'
    }
    register(
        id='FrankaCloth-v1',
        entry_point='pyrfuniverse.envs.robotics:FrankaClothEnv',
        kwargs=franka_cloth_kwargs,
        max_episode_steps=50
    )

    # Franka-Softbody env
    franka_softbody_kwargs = {
        'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/softbody',
        # 'executable_file': rfuniverse_build_base_root + 'FrankaSoftbodyServer/RFUniverse.x86_64',
        'executable_file': None,
        'reward_type': 'sparse'
    }
    register(
        id='FrankaSoftbody-v1',
        entry_point='pyrfuniverse.envs.robotics:FrankaSoftbodyEnv',
        kwargs=franka_softbody_kwargs,
        max_episode_steps=50
    )

    # Franka-ClothFold env
    franka_cloth_fold_kwargs = {
        'asset_bundle_file': '/home/haoyuan/rfuniverse/RFUniverse/Assets/AssetBundles/Linux/obi_cloth',
        # 'executable_file': rfuniverse_build_base_root + 'FrankaClothFoldServer/RFUniverse.x86_64',
        'executable_file': None,
        'reward_type': 'sparse'
    }
    register(
        id='FrankaClothFold-v1',
        entry_point='pyrfuniverse.envs.robotics:FrankaClothFoldEnv',
        kwargs=franka_cloth_fold_kwargs,
        max_episode_steps=20
    )

    # A relatively easy and naive NailCard environment
    register(
        id='NailCard-v1',
        entry_point='pyrfuniverse.envs:NailCardEnv',
        kwargs={
            'executable_file': rfuniverse_build_base_root + 'NailCardServer/RFUniverse.x86_64',
            'rotation_factor': 0,
            'goal_baseline': 0.02,
        },
        max_episode_steps=50,
    )

    # The normal version
    register(
        id='NailCard-v2',
        entry_point='pyrfuniverse.envs:NailCardEnv',
        kwargs={
            'executable_file': rfuniverse_build_base_root + 'NailCardServer/RFUniverse.x86_64',
        },
        max_episode_steps=50,
    )

    # Fixed the nail movement
    register(
        id='NailCard-v3',
        entry_point='pyrfuniverse.envs:NailCardEnv',
        kwargs={
            'executable_file': rfuniverse_build_base_root + 'NailCardServer/RFUniverse.x86_64',
            'nail_movement_factor': 0
        },
        max_episode_steps=50,
    )

    # A relatively easy and naive NailCard environment
    register(
        id='Robotiq85NailCard-v1',
        entry_point='pyrfuniverse.envs.gripper_nail:Robotiq85NailCardEnv',
        kwargs={
            'executable_file': rfuniverse_build_base_root + 'Robotiq85NailCardServer/RFUniverse.x86_64',
            'rotation_factor': 0,
            'goal_baseline': 0.02,
        },
        max_episode_steps=50,
    )

    # The normal version
    register(
        id='Robotiq85NailCard-v2',
        entry_point='pyrfuniverse.envs.gripper_nail:Robotiq85NailCardEnv',
        kwargs={
            'executable_file': rfuniverse_build_base_root + 'Robotiq85NailCardServer/RFUniverse.x86_64',
        },
        max_episode_steps=50,
    )

    # Fixed the nail movement
    register(
        id='Robotiq85NailCard-v3',
        entry_point='pyrfuniverse.envs.gripper_nail:Robotiq85NailCardEnv',
        kwargs={
            'executable_file': rfuniverse_build_base_root + 'Robotiq85NailCardServer/RFUniverse.x86_64',
            'vertical_movement_factor': 0.02,
            'nail_movement_factor': 0
        },
        max_episode_steps=50,
    )

    # Robotiq85 Nail Coin
    register(
        id='Robotiq85NailCoin-v1',
        entry_point='pyrfuniverse.envs.gripper_nail:Robotiq85NailCoinEnv',
        kwargs={
            'executable_file': rfuniverse_build_base_root + 'Robotiq85NailCoinServer/RFUniverse.x86_64',
            'vertical_movement_factor': 0.02,
            'nail_movement_factor': 0
        },
        max_episode_steps=50,
    )

    # Nail-Can environment
    register(
        id='NailCan-v1',
        entry_point='pyrfuniverse.envs:NailCanEnv',
        kwargs={
            'executable_file': rfuniverse_build_base_root + 'NailCanServer/RFUniverse.x86_64',
        }
    )

    register(
        id='Robotiq85NailCan-v1',
        entry_point='pyrfuniverse.envs.gripper_nail:Robotiq85NailCanEnv',
        kwargs={
            'executable_file': rfuniverse_build_base_root + 'Robotiq85NailCanServer/RFUniverse.x86_64',
        }
    )

    register(
        id='Robotiq85NailBook-v1',
        entry_point='pyrfuniverse.envs.gripper_nail:Robotiq85NailBookEnv',
        kwargs={
            'executable_file': rfuniverse_build_base_root + 'Robotiq85NailBook/RFUniverse.x86_64',
        }
    )

    # Robotiq85 insert
    register(
        id='Robotiq85Insert-v1',
        entry_point='pyrfuniverse.envs.gripper_nail:Robotiq85InsertEnv',
        kwargs={
            'executable_file': rfuniverse_build_base_root + 'Robotiq85InsertServer/RFUniverse.x86_64',
        }
    )

    register(
        id='Cleaner-v1',
        entry_point='pyrfuniverse.envs.multi_agent:CleanerEnv',
        kwargs={
            'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/rigidbody',
            'executable_file': rfuniverse_build_base_root + 'CleanerServer/RFUniverse.x86_64'
        }
    )

    register(
        id='Cleaner-v2',
        entry_point='pyrfuniverse.envs.multi_agent:CleanerEnv',
        kwargs={
            'reset_on_collision': False,
            'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/rigidbody',
            'executable_file': rfuniverse_build_base_root + 'CleanerServer/RFUniverse.x86_64',
        }
    )

    register(
        id='Cleaner-v3',
        entry_point='pyrfuniverse.envs.multi_agent:CleanerEnv',
        kwargs={
            'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/rigidbody',
            'executable_file': rfuniverse_build_base_root + 'CleanerWithWallServer/RFUniverse.x86_64'
        }
    )

    register(
        id='Cleaner-v4',
        entry_point='pyrfuniverse.envs.multi_agent:CleanerEnv',
        kwargs={
            'reset_on_collision': False,
            'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/rigidbody',
            'executable_file': rfuniverse_build_base_root + 'CleanerWithWallServer/RFUniverse.x86_64'
        }
    )

    register(
        id='Cleaner-v5',
        entry_point='pyrfuniverse.envs.multi_agent:CleanerEnv',
        kwargs={
            'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/rigidbody',
            'executable_file': rfuniverse_build_base_root + 'CleanerWithWallServer/RFUniverse.x86_64',
            'velocity_reward': True
        }
    )

    register(
        id='Cleaner-v6',
        entry_point='pyrfuniverse.envs.multi_agent:CleanerEnv',
        kwargs={
            'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/rigidbody',
            'executable_file': rfuniverse_build_base_root + 'CleanerWithWallServer/RFUniverse.x86_64',
            'velocity_reward': True,
            'collision_multiplier': 50,
        }
    )

    register(
        id='Cleaner-v7',
        entry_point='pyrfuniverse.envs.multi_agent:CleanerEnv',
        kwargs={
            'asset_bundle_file': '/home/haoyuan/rfuniverse/RFUniverse/Assets/AssetBundles/Linux/rigidbody',
            # 'executable_file': rfuniverse_build_base_root + 'CleanerWithWall/RFUniverse.x86_64',
            'executable_file': None,
            'velocity_reward': True,
            'collision_multiplier': 50,
            'obs_type': 'box',
            'max_episode_length': 200
        }
    )

    register(
        id='Cleaner-v8',
        entry_point='pyrfuniverse.envs.multi_agent:CleanerEnv',
        kwargs={
            'asset_bundle_file': rfuniverse_build_base_root + 'AssetBundles/rigidbody',
            'executable_file': rfuniverse_build_base_root + 'CleanerWithWallServer/RFUniverse.x86_64',
            'grid_reward_per_step': True
        }
    )

    register(
        id='UR5Box-v1',
        entry_point='pyrfuniverse.envs:Ur5BoxEnv',
        kwargs={
            'max_steps': 50,
            'reward_type': 'sparse',
            'min_open_angle': 30,
            'executable_file': rfuniverse_build_base_root + 'UR5Box/RFUniverse.x86_64'
        },
        max_episode_steps=50,
    )

    register(
        id='ToborPull-v1',
        entry_point='pyrfuniverse.envs.tobor_robotics:ToborPushPullEnv',
        kwargs={
            'max_steps': 100,
            'asset_bundle_file': '/home/haoyuan/workspace/rfuniverse/rfuniverse/RFUniverse/Assets/AssetBundles/Linux/articulation',
            'pull': True,
            'executable_file': rfuniverse_build_base_root + 'ToborPushPull/RFUniverse.x86_64'
        },
        max_episode_steps=100
    )

    register(
        id='ToborPush-v1',
        entry_point='pyrfuniverse.envs.tobor_robotics:ToborPushPullEnv',
        kwargs={
            'max_steps': 100,
            'asset_bundle_file': '/home/haoyuan/workspace/rfuniverse/rfuniverse/RFUniverse/Assets/AssetBundles/Linux/articulation',
            'pull': False,
            'executable_file': rfuniverse_build_base_root + 'ToborPushPull/RFUniverse.x86_64'
        },
        max_episode_steps=100
    )

except ImportError:
    print('No gym installed. Please install gym!')
    pass



from pyrfuniverse.envs.base_env import RFUniverseBaseEnv

env = RFUniverseBaseEnv(scene_file='HumanBodyIK.json')
env.step()
human = env.GetAttr(168242)
for index in range(5):
    human.HumanIKTargetDoMove(index=index,
                              position=[0, 0, 0.5],
                              duration=1,
                              speed_based=False,
                              relative=True
                              )
    env.step()
    human.WaitDo()
    human.HumanIKTargetDoMove(index=index,
                              position=[0, 0.5, 0],
                              duration=1,
                              speed_based=False,
                              relative=True
                              )
    env.step()
    human.WaitDo()
    human.HumanIKTargetDoMove(index=index,
                              position=[0, 0, -0.5],
                              duration=1,
                              speed_based=False,
                              relative=True
                              )
    env.step()
    human.WaitDo()
    human.HumanIKTargetDoMove(index=index,
                              position=[0, -0.5, 0],
                              duration=1,
                              speed_based=False,
                              relative=True
                              )
    env.step()
    human.WaitDo()

env.Pend()
env.close()



from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
import pyrfuniverse.attributes as attr
try:
    import pyrfuniverse.attributes.omplmanager_attr as rfu_ompl
except ImportError:
    raise Exception('This feature requires ompl, see: https://github.com/ompl/ompl')

env = RFUniverseBaseEnv(assets=['franka_panda', 'Collider_Box', 'OmplManager'])

robot = env.InstanceObject(name='franka_panda', id=123456, attr_type=attr.ControllerAttr)
robot.EnabledNativeIK(False)
box1 = env.InstanceObject(name='Collider_Box', id=111111, attr_type=attr.ColliderAttr)
box1.SetTransform(position=[-0.5, 0.5, 0], scale=[0.2, 1, 0.2])
box2 = env.InstanceObject(name='Collider_Box', id=111112, attr_type=attr.ColliderAttr)
box2.SetTransform(position=[0.5, 0.5, 0], scale=[0.2, 1, 0.2])
env.step()

ompl_manager = env.InstanceObject(name='OmplManager', attr_type=attr.OmplManagerAttr)
ompl_manager.modify_robot(123456)
env.step()

start_state = [0.0, -45.0, 0.0, -135.0, 0.0, 90.0, 45.0]
# target_state = [ompl_manager.joint_upper_limit[j] * 0.9 for j in range(ompl_manager.joint_num)]
target_state = [6.042808, -35.73029, -128.298, -118.3777, -40.28789, 134.8007, -139.2552]

planner = rfu_ompl.RFUOMPL(ompl_manager, time_unit=5)

ompl_manager.set_state(start_state)
env.step(50)

ompl_manager.set_state(target_state)
env.step(50)

ompl_manager.set_state(start_state)
env.step(50)

env.SetTimeStep(0.001)
env.step()

is_sol, path = planner.plan_start_goal(start_state, target_state)

print(target_state)
print(path[-1])

env.SetTimeStep(0.02)
env.step()

while True:
    if is_sol:
        planner.execute(path)


import os
os.environ["OPENCV_IO_ENABLE_OPENEXR"] = "1"
import cv2
import numpy as np
try:
    import open3d as o3d
except ImportError:
    raise Exception('This feature requires open3d, please install with `pip install open3d`')
import pyrfuniverse.utils.rfuniverse_utility as utility
import pyrfuniverse.utils.depth_processor as dp
from pyrfuniverse.envs.base_env import RFUniverseBaseEnv


nd_main_intrinsic_matrix = np.array([[600, 0, 240],
                                     [0, 600, 240],
                                     [0, 0, 1]])
nd_ir_intrinsic_matrix = np.array([[480, 0, 240],
                                   [0, 480, 240],
                                   [0, 0, 1]])

env = RFUniverseBaseEnv(scene_file='ActiveDepth.json')
active_light_sensor_1 = env.GetAttr(789789)

active_light_sensor_1.GetRGB(intrinsic_matrix=nd_main_intrinsic_matrix)
env.step()
image_byte = active_light_sensor_1.data['rgb']
image_rgb = np.frombuffer(image_byte, dtype=np.uint8)
image_rgb = cv2.imdecode(image_rgb, cv2.IMREAD_COLOR)
image_rgb = cv2.cvtColor(image_rgb, cv2.COLOR_BGR2RGB)
image_rgb = np.transpose(image_rgb, [1, 0, 2])

active_light_sensor_1.GetID(intrinsic_matrix=nd_main_intrinsic_matrix)
env.step()
image_id = active_light_sensor_1.data['id_map']
image_id = np.frombuffer(image_id, dtype=np.uint8)
image_id = cv2.imdecode(image_id, cv2.IMREAD_COLOR)
image_id = cv2.cvtColor(image_id, cv2.COLOR_BGR2RGB)

active_light_sensor_1.GetDepthEXR(intrinsic_matrix=nd_main_intrinsic_matrix)
env.step()
image_depth_exr = active_light_sensor_1.data['depth_exr']

active_light_sensor_1.GetActiveDepth(main_intrinsic_matrix_local=nd_main_intrinsic_matrix,
                                     ir_intrinsic_matrix_local=nd_ir_intrinsic_matrix)
env.step()
image_active_depth = active_light_sensor_1.data['active_depth']
image_active_depth = np.transpose(image_active_depth, [1, 0])

local_to_world_matrix = active_light_sensor_1.data['local_to_world_matrix']
# local_to_world_matrix = np.reshape(local_to_world_matrix, [4, 4]).T

# point = dp.image_array_to_point_cloud(image_rgb, image_active_depth, 45, local_to_world_matrix)

color = utility.EncodeIDAsColor(568451)[0:3]
real_point_cloud1 = dp.image_bytes_to_point_cloud_intrinsic_matrix(image_byte, image_depth_exr, nd_main_intrinsic_matrix, local_to_world_matrix)
real_point_cloud1 = dp.mask_point_cloud_with_id_color(real_point_cloud1, image_id, color)
active_point_cloud1 = dp.image_array_to_point_cloud_intrinsic_matrix(image_rgb, image_active_depth, nd_main_intrinsic_matrix, local_to_world_matrix)
active_point_cloud1 = dp.mask_point_cloud_with_id_color(active_point_cloud1, image_id, color)
filtered_point_cloud1 = dp.filter_active_depth_point_cloud_with_exact_depth_point_cloud(active_point_cloud1, real_point_cloud1)

##################################################

active_light_sensor_2 = env.GetAttr(123123)

active_light_sensor_2.GetRGB(intrinsic_matrix=nd_main_intrinsic_matrix)
env.step()
image_byte = active_light_sensor_2.data['rgb']
image_rgb = np.frombuffer(image_byte, dtype=np.uint8)
image_rgb = cv2.imdecode(image_rgb, cv2.IMREAD_COLOR)
image_rgb = cv2.cvtColor(image_rgb, cv2.COLOR_BGR2RGB)
image_rgb = np.transpose(image_rgb, [1, 0, 2])

active_light_sensor_2.GetID(intrinsic_matrix=nd_main_intrinsic_matrix)
env.step()
image_id = active_light_sensor_2.data['id_map']
image_id = np.frombuffer(image_id, dtype=np.uint8)
image_id = cv2.imdecode(image_id, cv2.IMREAD_COLOR)
image_id = cv2.cvtColor(image_id, cv2.COLOR_BGR2RGB)

active_light_sensor_2.GetDepthEXR(intrinsic_matrix=nd_main_intrinsic_matrix)
env.step()
image_depth_exr = active_light_sensor_2.data['depth_exr']

active_light_sensor_2.GetActiveDepth(main_intrinsic_matrix_local=nd_main_intrinsic_matrix,
                                     ir_intrinsic_matrix_local=nd_ir_intrinsic_matrix)
env.step()
image_active_depth = active_light_sensor_2.data['active_depth']
image_active_depth = np.transpose(image_active_depth, [1, 0])

local_to_world_matrix = active_light_sensor_2.data['local_to_world_matrix']

env.close()

# point = dp.image_array_to_point_cloud(image_rgb, image_active_depth, 45, local_to_world_matrix)
real_point_cloud2 = dp.image_bytes_to_point_cloud_intrinsic_matrix(image_byte, image_depth_exr,
                                                                   nd_main_intrinsic_matrix, local_to_world_matrix)
real_point_cloud2 = dp.mask_point_cloud_with_id_color(real_point_cloud2, image_id, color)
active_point_cloud2 = dp.image_array_to_point_cloud_intrinsic_matrix(image_rgb, image_active_depth,
                                                                     nd_main_intrinsic_matrix, local_to_world_matrix)
active_point_cloud2 = dp.mask_point_cloud_with_id_color(active_point_cloud2, image_id, color)
filtered_point_cloud2 = dp.filter_active_depth_point_cloud_with_exact_depth_point_cloud(active_point_cloud2, real_point_cloud2)

# unity space to open3d space and show
real_point_cloud1.transform([[-1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
real_point_cloud2.transform([[-1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
active_point_cloud1.transform([[-1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
active_point_cloud2.transform([[-1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
filtered_point_cloud1.transform([[-1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
filtered_point_cloud2.transform([[-1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
coorninate = o3d.geometry.TriangleMesh.create_coordinate_frame()

o3d.visualization.draw_geometries([real_point_cloud1, real_point_cloud2, coorninate])
o3d.visualization.draw_geometries([active_point_cloud1, active_point_cloud2, coorninate])
o3d.visualization.draw_geometries([filtered_point_cloud1, filtered_point_cloud2, coorninate])



from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
import pyrfuniverse.attributes as attr

env = RFUniverseBaseEnv()

digit = env.InstanceObject(name='Digit', attr_type=attr.DigitAttr)
digit.SetTransform(position=[0, 0.015, 0])
target = env.InstanceObject(name='DigitTarget')
target.SetTransform(position=[0, 0.05, 0.015])
env.SetViewTransform(position=[-0.1, 0.033, 0.014], rotation=[0, 90, 0])
env.Pend()
env.close()




import os
from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
import pyrfuniverse.attributes as attr

# supported gripper: allegro_hand_right, bhand, dh_robotics_ag95_gripper, franka_hand, svh
env = RFUniverseBaseEnv(assets=['allegro_hand_right'])
bhand = env.InstanceObject("allegro_hand_right", attr_type=attr.ControllerAttr)
env.step(5)
moveable_joint_count = bhand.data['number_of_moveable_joints']
print(f'moveable_joint_count:{moveable_joint_count}')
bhand.SetJointPositionDirectly([30 for _ in range(moveable_joint_count)])
env.step(5)
env.ExportOBJ([bhand.id], os.path.abspath('../Mesh/gripper_mesh.obj'))
env.Pend()
env.close()




import pyrfuniverse.attributes as attr


class CustomAttr(attr.BaseAttr):
    """
    This is an example of custom attribute class, without actual functions.
    """
    def parse_message(self, data: dict):
        """
        Parse messages. This function is called by internal function.

        Returns:
            Dict: A dict containing useful information of this class.

            data['custom_message']: A custom message
        """
        super().parse_message(data)

    # An example of new API
    def CustomMessage(self, message: str):
        self._send_data('CustomMessage', message)



import math
import os
import os.path as osp
import random

import numpy as np
import open3d as o3d
import cv2
import tempfile

def image_bytes_to_point_cloud(rgb_bytes: bytes, depth_bytes: bytes, fov: float, local_to_world_matrix: np.ndarray):
    """
    Use the raw bytes of RGB image and depth image, as well as the camera
    FOV and extrinsic matrix to generate point cloud in global coordinate.

    Args:
        rgb_bytes: Bytes, raw bytes of RGB image.
        depth_bytes: Bytes, raw bytes of depth image.
        fov: Float, camera Field Of View (FOV).
        local_to_world_matrix: Numpy.ndarray, the local_to_world_matrix of camera.
    
    Return:
        open3d.geometry.PointCloud: The point cloud.
    """
    image_rgb = np.frombuffer(rgb_bytes, dtype=np.uint8)
    image_rgb = cv2.imdecode(image_rgb, cv2.IMREAD_COLOR)
    image_rgb = cv2.cvtColor(image_rgb, cv2.COLOR_BGR2RGB)
    image_rgb = np.transpose(image_rgb, [1, 0, 2])

    temp_file_path = osp.join(tempfile.gettempdir(), 'temp_img.exr')
    with open(temp_file_path, 'wb') as f:
        f.write(depth_bytes)
    image_depth = cv2.imread(temp_file_path, cv2.IMREAD_UNCHANGED)
    os.remove(temp_file_path)

    image_depth = np.transpose(image_depth, [1, 0])

    pcd = image_array_to_point_cloud(image_rgb, image_depth, fov, local_to_world_matrix)
    return pcd


def image_array_to_point_cloud(image_rgb: np.ndarray, image_depth: np.ndarray, fov: float, local_to_world_matrix: np.ndarray):
    """
    Use the RGB image and depth image, as well as the camera
    FOV and extrinsic matrix to generate point cloud in global coordinate.

    Args:
        image_rgb: Numpy.ndarray, in shape (H,W,3), the RGB image.
        image_depth: Numpy.ndarray, in shape (H,W,3), the depth image.
        fov: Float, camera Field Of View (FOV).
        local_to_world_matrix: Numpy.ndarray, the local_to_world_matrix of camera.
    
    Return:
        open3d.geometry.PointCloud: The point cloud.
    """
    points = depth_to_point_cloud(image_depth, fov=fov, organized=False)
    pcd = o3d.geometry.PointCloud()
    pcd.points = o3d.utility.Vector3dVector(points)

    colors = image_rgb.reshape(-1, 3)
    pcd.colors = o3d.utility.Vector3dVector(colors.astype(np.float32) / 255.0)

    pcd.transform(local_to_world_matrix)

    return pcd

def depth_to_point_cloud(depth: np.ndarray, fov: float, organized=False):
    """
    Use the depth image and the camera FOV to generate point cloud in
    camera coordinate.

    Args:
        depth: Numpy.ndarray, in shape (H,W,3), the depth image.
        fov: Float, camera Field Of View (FOV).
        organized: Bool, whether keep the point organized. If True, the returned shape of point cloud is (H,W,3); If False, the returned shape of point cloud is (H*W,3).
    
    Return:
        open3d.geometry.PointCloud: The point cloud.
    """
    width = depth.shape[0]
    height = depth.shape[1]
    cx = width / 2
    cy = height / 2
    xmap = np.arange(width)
    ymap = np.arange(height)
    xmap, ymap = np.meshgrid(xmap, ymap)
    xmap = xmap.T
    ymap = ymap.T

    fx = fy = height / (2 * math.tan(math.radians(fov / 2)))

    points_z = depth
    points_x = (xmap - cx) * points_z / fx
    points_y = (ymap - cy) * points_z / fy

    # radian_per_pixel = math.radians(fov / height)
    # points_x = np.sin(radian_per_pixel * (xmap - cx)) * depth
    # points_y = np.sin(radian_per_pixel * (ymap - cy)) * depth
    # points_z = depth ** 2 - points_x ** 2 - points_y ** 2
    # points_z[points_z < 0] = 0
    # points_z = np.sqrt(points_z)

    cloud = np.stack([points_x, -points_y, points_z], axis=-1)
    if not organized:
        cloud = cloud.reshape([-1, 3])

    return cloud

def image_bytes_to_point_cloud_intrinsic_matrix(rgb_bytes: bytes, depth_bytes: bytes, intrinsic_matrix: np.ndarray, local_to_world_matrix: np.ndarray):
    """
    Use the raw bytes of RGB image and depth image, as well as the camera
    intrinsic matrix and extrinsic matrix to generate point cloud in global coordinate.

    Args:
        rgb_bytes: Bytes, raw bytes of RGB image.
        depth_bytes: Bytes, raw bytes of depth image.
        intrinsic_matrix: Numpy.ndarray, the intrinsic matrix of camera.
        local_to_world_matrix: Numpy.ndarray, the local_to_world_matrix of camera.
    
    Return:
        open3d.geometry.PointCloud: The point cloud.
    """
    temp_file_path = osp.join(tempfile.gettempdir(), f'temp_img_{int(random.uniform(10000000,99999999))}.png')
    with open(temp_file_path, 'wb') as f:
        f.write(rgb_bytes)
    color = o3d.io.read_image(temp_file_path)
    os.remove(temp_file_path)

    temp_file_path = osp.join(tempfile.gettempdir(), f'temp_img_{int(random.uniform(10000000,99999999))}.exr')
    with open(temp_file_path, 'wb') as f:
        f.write(depth_bytes)
    # change .exr format to .png format
    depth_exr = cv2.imread(temp_file_path, cv2.IMREAD_UNCHANGED)
    os.remove(temp_file_path)
    # mask = np.asarray(o3d.io.read_image(osp.join(mask_path, video_name, view_name, base_name + '.png')))[:, :, 0]
    # foregound_mask = mask == 11
    depth_png = (depth_exr * 1000).astype(np.uint16)[:, :]
    # foreground_depth_png[~foregound_mask] = 0  # filter the background, only need foreground
    temp_file_path = osp.join(tempfile.gettempdir(), f'temp_img_{int(random.uniform(10000000,99999999))}.png')
    cv2.imwrite(temp_file_path, depth_png)
    depth = o3d.io.read_image(temp_file_path)
    os.remove(temp_file_path)

    pcd = image_open3d_to_point_cloud_intrinsic_matrix(color, depth, intrinsic_matrix, local_to_world_matrix)

    return pcd


def image_array_to_point_cloud_intrinsic_matrix(image_rgb: np.ndarray, image_depth: np.ndarray, intrinsic_matrix: np.ndarray, local_to_world_matrix: np.ndarray):
    """
    Use the RGB image and depth image, as well as the camera
    intrinsic matrix and extrinsic matrix to generate point cloud in global coordinate.

    Args:
        image_rgb: Numpy.ndarray, in shape (H,W,3), the RGB image.
        image_depth: Numpy.ndarray, in shape (H,W,3), the depth image.
        intrinsic_matrix: Numpy.ndarray, the intrinsic matrix of camera.
        local_to_world_matrix: Numpy.ndarray, the local_to_world_matrix of camera.
    
    Return:
        open3d.geometry.PointCloud: The point cloud.
    """
    temp_file_path = osp.join(tempfile.gettempdir(), f'temp_img_{int(random.uniform(10000000,99999999))}.png')

    image_rgb = np.transpose(image_rgb, [1, 0, 2])
    image_rgb = cv2.cvtColor(image_rgb, cv2.COLOR_RGB2BGR)
    cv2.imwrite(temp_file_path, image_rgb)
    color = o3d.io.read_image(temp_file_path)

    # change .exr format to .png format
    depth_png = (image_depth * 1000).astype(np.uint16)[:, :]
    depth_png = np.transpose(depth_png, [1, 0])
    cv2.imwrite(temp_file_path, depth_png)
    depth = o3d.io.read_image(temp_file_path)

    os.remove(temp_file_path)

    pcd = image_open3d_to_point_cloud_intrinsic_matrix(color, depth, intrinsic_matrix, local_to_world_matrix)

    return pcd

def image_open3d_to_point_cloud_intrinsic_matrix(color: o3d.geometry.Image, depth: o3d.geometry.Image, intrinsic_matrix: np.ndarray, local_to_world_matrix: np.ndarray):
    """
    Use the RGB image and depth image in open3d.geometry.Image format, as well as the camera
    intrinsic matrix and extrinsic matrix to generate point cloud in global coordinate.

    Args:
        color: open3d.geometry.Image, the RGB image.
        depth: open3d.geometry.Image, the depth image.
        intrinsic_matrix: Numpy.ndarray, the intrinsic matrix of camera.
        local_to_world_matrix: Numpy.ndarray, the local_to_world_matrix of camera.
    
    Return:
        open3d.geometry.PointCloud: The point cloud.
    """
    rgbd_image = o3d.geometry.RGBDImage.create_from_color_and_depth(
        color, depth, depth_trunc=20, convert_rgb_to_intensity=False)

    intrinsic_matrix = o3d.camera.PinholeCameraIntrinsic(int(intrinsic_matrix[0, 2] * 2),
                                                         int(intrinsic_matrix[1, 2] * 2), intrinsic_matrix[0, 0],
                                                         intrinsic_matrix[1, 1], intrinsic_matrix[0, 2],
                                                         intrinsic_matrix[1, 2])

    pcd = o3d.geometry.PointCloud.create_from_rgbd_image(
        rgbd_image, intrinsic_matrix, project_valid_depth_only=False)

    # convter to unity space
    pcd.transform([[1, 0, 0, 0], [0, -1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])

    # camera to world in unity space
    pcd.transform(local_to_world_matrix)

    return pcd

def mask_point_cloud_with_id_color(pcd: o3d.geometry.PointCloud, image_mask: np.ndarray, color:list):
    """
    Mask the point cloud with given segmentation masks and target color.

    Args:
        pcd: open3d.geometry.PointCloud, the point cloud.
        image_mask: numpy.ndarray, the segmentation mask in shape (H,W,3).
        color: List, the target color list.
    
    Return:
        open3d.geometry.PointCloud: The point cloud.
    """
    image_mask = image_mask.reshape(-1, 3)
    index = np.argwhere(image_mask == color)[:, 0]
    index = index[::3]
    d = np.array(pcd.points)[index]
    pcd.points = o3d.utility.Vector3dVector(d)
    c = np.array(pcd.colors)[index]
    pcd.colors = o3d.utility.Vector3dVector(c)
    return pcd

def mask_point_cloud_with_id_gray_color(pcd: o3d.geometry.PointCloud, image_mask: np.ndarray, color:int):
    """
    Mask the point cloud with given gray-scale segmentation masks and target color.

    Args:
        pcd: open3d.geometry.PointCloud, the point cloud.
        image_mask: numpy.ndarray, the segmentation mask in shape (H,W).
        color: Int, the target gray-scale color.
    
    Return:
        open3d.geometry.PointCloud: The point cloud.
    """
    image_mask = image_mask.reshape(-1)
    index = np.argwhere(image_mask == color).reshape(-1)
    id_pcd = o3d.geometry.PointCloud()
    d = np.array(pcd.points)[index]
    id_pcd.points = o3d.utility.Vector3dVector(d)
    c = np.array(pcd.colors)[index]
    id_pcd.colors = o3d.utility.Vector3dVector(c)
    return id_pcd

def filter_active_depth_point_cloud_with_exact_depth_point_cloud(active_pcd: o3d.geometry.PointCloud, exact_pcd: o3d.geometry.PointCloud, max_distance:float = 0.05):
    """
    Use exact point cloud to filter IR-based active point cloud based on a tolerance distance.

    Args:
        active_pcd: open3d.geometry.PointCloud, the IR-based active point cloud.
        exact_pcd: open3d.geometry.PointCloud, the exact point cloud.
        max_distance: float, the maximum tolerance distance.
    
    Return:
        open3d.geometry.PointCloud: The point cloud.
    """
    active_point = np.array(active_pcd.points)
    exact_point = np.array(exact_pcd.points)
    # m = exact_point.shape[0]
    # n = active_point.shape[0]
    # duplicated_active_point = np.repeat(active_point, m).reshape((n, m, 3))
    # distance = np.linalg.norm(duplicated_active_point - exact_point, axis=-1)
    distance = np.linalg.norm(active_point - exact_point, axis=-1)
    # min_distance = np.min(distance, axis=1)
    index = np.argwhere(distance < max_distance).reshape(-1)
    filter_pcd = o3d.geometry.PointCloud()
    d = np.array(active_pcd.points)[index]
    filter_pcd.points = o3d.utility.Vector3dVector(d)
    c = np.array(active_pcd.colors)[index]
    filter_pcd.colors = o3d.utility.Vector3dVector(c)
    return filter_pcd



from pyrfuniverse.envs.robotics.franka_robotics_env import FrankaRoboticsEnv
from pyrfuniverse.envs.robotics.pick_and_place_env import FrankaPickAndPlaceEnv
from pyrfuniverse.envs.robotics.push_env import FrankaPushEnv
from pyrfuniverse.envs.robotics.reach_env import FrankaReachEnv
from pyrfuniverse.envs.robotics.franka_cloth_env import FrankaClothEnv
from pyrfuniverse.envs.robotics.franka_softbody_env import FrankaSoftbodyEnv
from pyrfuniverse.envs.robotics.franka_cloth_fold_env import FrankaClothFoldEnv

__all__ = [
    'FrankaRoboticsEnv', 'FrankaReachEnv', 'FrankaPushEnv', 'FrankaPickAndPlaceEnv',
    'FrankaClothEnv', 'FrankaSoftbodyEnv', 'FrankaClothFoldEnv'
]



import pybullet as p
import pybullet_data
import numpy as np
import math


class RFUniverseJacoController:
    def __init__(
            self,
            robot_urdf,
            base_pos=np.array([0, 0, 0]),
            base_orn=[-0.707107, 0.0, 0.0, 0.707107],
            init_joint_positions=[0, 47, 0, 30, 0, 65, 0, 0, 0, 0],
            render=False
    ):
        if render:
            p.connect(p.GUI)  # For debug mode
        else:
            p.connect(p.DIRECT)

        p.configureDebugVisualizer(p.COV_ENABLE_Y_AXIS_UP, 1)
        p.setAdditionalSearchPath(pybullet_data.getDataPath())
        p.setGravity(0, -9.8, 0)

        self.bullet_client = p
        self.bullet_flags = self.bullet_client.URDF_ENABLE_CACHED_GRAPHICS_SHAPES

        self.robot_name = 'jaco'
        self.robot_urdf = robot_urdf
        self.end_effector_id = 8
        self.num_dof = 10
        self.init_joint_positions = self.get_pybullet_joint_pos_from_unity(init_joint_positions)
        self.bullet_base_pos = np.array(self.get_bullet_pos_from_unity(base_pos))
        self.robot = self.bullet_client.loadURDF(self.robot_urdf, self.bullet_base_pos, base_orn, useFixedBase=True,
                                                 flags=self.bullet_flags)

        self.revolute_joint_ids = []
        for j in range(15):
            self.bullet_client.changeDynamics(self.robot, j, linearDamping=0, angularDamping=0)
            info = self.bullet_client.getJointInfo(self.robot, j)

            jointName = info[1]
            jointType = info[2]
            # print(j, jointName)
            if (jointType == self.bullet_client.JOINT_REVOLUTE):
                self.revolute_joint_ids.append(j)

        self.reset()

        # print('pybullet', self.get_link_state(self.end_effector_id))

    def get_bullet_pos_from_unity(self, unity_pos: list) -> list:
        return [-1 * unity_pos[0], unity_pos[1], unity_pos[2]]

    def get_unity_pos_from_bullet(self, bullet_pos: list) -> list:
        return [-1 * bullet_pos[0], bullet_pos[1], bullet_pos[2]]

    def get_unity_joint_pos_from_pybullet(self, pybullet_joint_pos: tuple) -> list:
        pybullet_joint_pos = list(pybullet_joint_pos)[:self.num_dof]
        for i, (joint_pos) in enumerate(pybullet_joint_pos):
            pybullet_joint_pos[i] = 180 * joint_pos / math.pi

        return pybullet_joint_pos

    def get_pybullet_joint_pos_from_unity(self, unity_joint_pos: list) -> list:
        unity_joint_pos = list(unity_joint_pos)[:self.num_dof]
        unity_joint_pos = np.array(unity_joint_pos)
        pybullet_joint_pos = unity_joint_pos * math.pi / 180

        return pybullet_joint_pos

    def calculate_ik(self, unity_eef_pos, eef_orn=None) -> list:
        if eef_orn is None:
            eef_orn = self.bullet_client.getQuaternionFromEuler([math.pi / 2., 0., 0.])

        eef_pos = self.get_bullet_pos_from_unity(unity_eef_pos)

        joint_positions = self.bullet_client.calculateInverseKinematics(
            self.robot, self.end_effector_id, eef_pos, eef_orn, maxNumIterations=20)

        for i, (idx) in enumerate(self.revolute_joint_ids):
            self.bullet_client.resetJointState(self.robot, idx, joint_positions[i])

        return self.get_unity_joint_pos_from_pybullet(joint_positions)

    def calculate_ik_recursive(self, unity_eef_pos, eef_orn=None) -> list:
        if eef_orn is None:
            eef_orn = self.bullet_client.getQuaternionFromEuler([math.pi / 2., 0., 0.])

        eef_pos = self.get_bullet_pos_from_unity(unity_eef_pos)
        for i in range(20):
            joint_positions = self.bullet_client.calculateInverseKinematics(
                self.robot, self.end_effector_id, eef_pos, eef_orn, maxNumIterations=20)

            for i, (idx) in enumerate(self.revolute_joint_ids):
                self.bullet_client.resetJointState(self.robot, idx, joint_positions[i])

        return self.get_unity_joint_pos_from_pybullet(joint_positions)

    def get_link_state(self, link_idx):
        link_state = self.bullet_client.getLinkState(self.robot, link_idx)

        return self.get_unity_pos_from_bullet(link_state[0])

    def reset(self):
        for i, (idx) in enumerate(self.revolute_joint_ids):
            self.bullet_client.resetJointState(self.robot, idx, self.init_joint_positions[i])


'''
A OMPL plugin for RFUniverse
Modified from pybullet_ompl

Author: Jieyi Zhang
'''

INTERPOLATE_NUM = 300
DEFAULT_PLANNING_TIME = 500.0

from ompl import base as ob
from ompl import geometric as og
import copy
import pyrfuniverse.attributes as attr


class OmplManagerAttr(attr.BaseAttr):
    """
    To use with rfu_OMPL. You need to construct a instance of this class and pass to PbOMPL.

    Note:
    This parent class by default assumes that all joints are acutated and should be planned. If this is not your desired
    behaviour, please write your own inheritated class that overrides respective functionalities.
    """

    def __init__(self, env, id: int, data=None):
        super().__init__(env, id, data)
        self.is_collision = False
        self.state = None
        self.robot_attr = None
        self.joint_num = 0
        self.joint_lower_limit = []
        self.joint_upper_limit = []

    def parse_message(self, data: dict):
        super().parse_message(data)
        self.is_collision = data['is_collide']

    def modify_robot(self, robot_id: int):
        self.robot_attr = self.env.GetAttr(robot_id)
        self.joint_num = self.robot_attr.data['number_of_moveable_joints']
        self.joint_lower_limit = self.robot_attr.data['joint_lower_limit']
        self.joint_upper_limit = self.robot_attr.data['joint_upper_limit']

        self._send_data('ModifyRobot', robot_id)

    def get_cur_state(self) -> list:
        """
        Get current robot state.
        """
        return copy.deepcopy(self.state)

    def set_state(self, state):
        """
        Set robot state.
        To faciliate collision checking
        Args:
            state: list[Float], joint values of robot
        """
        self.state = state
        self._set_joint_positions(self.state)

    def reset(self):
        """
        Reset robot state
        Args:
            state: list[Float], joint values of robot
        """
        self.state = [0.0] * self.joint_num
        self._set_joint_positions(self.state)

    def _set_joint_positions(self, positions: list):
        self.is_collision = False

        self._send_data('SetJointState', positions)

    def RestoreRobot(self, robot_id: int):
        self._send_data('RestoreRobot', robot_id)


class RFUStateSpace(ob.RealVectorStateSpace):
    def __init__(self, num_dim) -> None:
        super().__init__(num_dim)
        self.num_dim = num_dim
        self.state_sampler = None

    def allocStateSampler(self):
        """
        This will be called by the internal OMPL planner
        """
        # WARN: This will cause problems if the underlying planner is multi-threaded!!!
        if self.state_sampler:
            return self.state_sampler

        # when ompl planner calls this, we will return our sampler
        return self.allocDefaultStateSampler()

    def set_state_sampler(self, state_sampler):
        """
        Optional, Set custom state sampler.
        """
        self.state_sampler = state_sampler


class RFUOMPL():
    def __init__(self, manager, time_unit=1) -> None:
        """
        Args
            robot: A RFUOMPLRobot instance.
            obstacles: list of obstacle ids. Optional.
        """
        self.manager = manager
        self.env = self.manager.env
        self.time_unit = time_unit

        self.space = RFUStateSpace(self.manager.joint_num)

        bounds = ob.RealVectorBounds(self.manager.joint_num)
        upper_bound = self.manager.joint_upper_limit
        lower_bound = self.manager.joint_lower_limit
        for i in range(self.manager.joint_num):
            bounds.setLow(i, lower_bound[i])
            bounds.setHigh(i, upper_bound[i])
        self.space.setBounds(bounds)

        self.ss = og.SimpleSetup(self.space)
        self.ss.setStateValidityChecker(ob.StateValidityCheckerFn(self.is_state_valid))
        self.si = self.ss.getSpaceInformation()
        # self.si.setStateValidityCheckingResolution(0.005)
        # self.collision_fn = pb_utils.get_collision_fn(self.robot_id, self.robot.joint_idx, self.obstacles, [], True, set(),
        #                                                 custom_limits={}, max_distance=0, allow_collision_links=[])

        self.set_planner("InformedRRTstar") # RRT by default

    def is_state_valid(self, state):
        # check if a given state will lead a collision
        self.manager.set_state(self.state_to_list(state))
        # self.env.SetNextStepNoTimeConsuming()
        self.env.step(2)
        return not self.manager.is_collision

    def set_planner(self, planner_name):
        """
        Note: Add your planner here!!
        """
        if planner_name == "PRM":
            self.planner = og.PRM(self.ss.getSpaceInformation())
        elif planner_name == "RRT":
            self.planner = og.RRT(self.ss.getSpaceInformation())
        elif planner_name == "RRTConnect":
            self.planner = og.RRTConnect(self.ss.getSpaceInformation())
        elif planner_name == "RRTstar":
            self.planner = og.RRTstar(self.ss.getSpaceInformation())
        elif planner_name == "EST":
            self.planner = og.EST(self.ss.getSpaceInformation())
        elif planner_name == "FMT":
            self.planner = og.FMT(self.ss.getSpaceInformation())
        elif planner_name == "BITstar":
            self.planner = og.BITstar(self.ss.getSpaceInformation())
        elif planner_name == "InformedRRTstar":
            self.planner = og.InformedRRTstar(self.ss.getSpaceInformation())
        else:
            print("{} not recognized, please add it first".format(planner_name))
            return
        self.ss.setPlanner(self.planner)

    def plan_start_goal(self, start, goal, until_success=True, allowed_time=None):
        """
        plan a path to gaol from the given robot start state
        """
        if allowed_time is None:
            allowed_time = self.time_unit

        print("start_planning")
        print(self.planner.params())

        orig_robot_state = self.manager.get_cur_state()

        # set the start and goal states;
        s = ob.State(self.space)
        g = ob.State(self.space)
        for i in range(len(start)):
            s[i] = start[i]
            g[i] = goal[i]

        # print(self.is_state_valid(goal))
        # print(g)
        # print(goal)
        self.ss.setStartAndGoalStates(s, g)

        # attempt to solve the problem within allowed planning time
        res = False
        self.ss.solve(allowed_time)
        print('Solution Find')
        while True:
            print("Found solution: interpolating into {} segments".format(INTERPOLATE_NUM))
            # print the path to screen
            sol_path_geometric = self.ss.getSolutionPath()
            sol_path_geometric.interpolate(INTERPOLATE_NUM)
            sol_path_states = sol_path_geometric.getStates()
            sol_path_list = [self.state_to_list(state) for state in sol_path_states]
            sum = 0
            last_state = sol_path_list[-1]
            for i in range(self.manager.joint_num):
                sum += abs(goal[i] - last_state[i])
            if sum < 1e-2:
                res = True
            if not until_success or res:
                break
            allowed_time *= 2
            self.ss.solve(allowed_time)

        # reset robot state
        self.manager.set_state(orig_robot_state)
        self.env.step(50)
        return res, sol_path_list

    def plan(self, goal, allowed_time = DEFAULT_PLANNING_TIME):
        """
        plan a path to gaol from current robot state
        """
        start = self.manager.get_cur_state()
        return self.plan_start_goal(start, goal, allowed_time=allowed_time)

    def execute(self, path, dynamics=False):
        """
        Execute a planned plan. Will visualize in pybullet.
        Args:
            path: list[state], a list of state
            dynamics: allow dynamic simulation. If dynamics is false, this API will use robot.set_state(),
                      meaning that the simulator will simply reset robot's state WITHOUT any dynamics simulation. Since the
                      path is collision free, this is somewhat acceptable.
        """
        for q in path:
            if dynamics:
                for i in range(self.manager.num_dim):
                    self.manager.control_attr.SetJointPosition(q)
                    self.env.step(100)
            else:
                self.manager.set_state(q)
            self.env.step()

    def set_state_sampler(self, state_sampler):
        self.space.set_state_sampler(state_sampler)

    def state_to_list(self, state):
        return [state[i] for i in range(self.manager.joint_num)]


from pyrfuniverse.envs.gym_goal_wrapper_env import RFUniverseGymGoalWrapper
# from pyrfuniverse.utils import RFUniverseController
import numpy as np
import math
import copy
from gym import spaces
from gym.utils import seeding


class FrankaSoftbodyEnv(RFUniverseGymGoalWrapper):
    metadata = {'render.modes': ['human']}

    def __init__(
            self,
            asset_bundle_file,
            reward_type,
            open_gripper=True,
            tolerance=0.05,
            goal_xz_range=0.15,
            softbody_xz_range=0.15,
            executable_file=None,
    ):
        super().__init__(
            executable_file=executable_file,
            articulation_channel=True,
            game_object_channel=True,
            obi_softbody_channel=True
        )
        self.asset_bundle_file = asset_bundle_file
        self.reward_type = reward_type
        self.open_gripper = open_gripper
        self.tolerance = tolerance

        # Fixed parameters
        self.scale = 10.0
        self.goal_center_position = np.array([-0.6, 0.015, 0])
        self.softbody_center_position = np.array([-0.6, 0.022, 0])
        self.goal_range_low = self.goal_center_position - np.array([goal_xz_range, 0, goal_xz_range])
        self.goal_range_high = self.goal_center_position + np.array([goal_xz_range, 0, goal_xz_range])
        self.softbody_range_low = self.softbody_center_position - np.array([softbody_xz_range, 0, softbody_xz_range])
        self.softbody_range_high = self.softbody_center_position + np.array([softbody_xz_range, 0, softbody_xz_range])

        # Env setup
        self.seed()
        self.load_softbody = False
        self._load_obi_softbody()
        self.ik_controller = RFUniverseController('franka', base_pos=np.array([0, 0, 0]))
        self.t = 0
        self.goal = self._sample_goal()
        self.action_space = spaces.Box(
            low=-1, high=1, shape=(4,), dtype=np.float32
        )
        obs = self._get_obs()
        self.observation_space = spaces.Dict({
            'observation': spaces.Box(-np.inf, np.inf, shape=obs['observation'].shape, dtype=np.float32),
            'desired_goal': spaces.Box(-np.inf, np.inf, shape=obs['desired_goal'].shape, dtype=np.float32),
            'achieved_goal': spaces.Box(-np.inf, np.inf, shape=obs['achieved_goal'].shape, dtype=np.float32)
        })

    def step(self, action: np.ndarray):
        pos_ctrl = action[:3] * 0.05
        curr_pos = self._get_gripper_position()
        pos_ctrl = curr_pos + pos_ctrl
        joint_positions = self.ik_controller.calculate_ik(pos_ctrl)

        if self.open_gripper:
            curr_gripper_width = self._get_gripper_width()
            target_gripper_width = curr_gripper_width + action[3] * 0.2
            target_gripper_width = np.clip(target_gripper_width, 0, 0.08)
            joint_positions.append(target_gripper_width)
        else:
            joint_positions.append(0)

        self._set_franka_joints(np.array(joint_positions))
        self.t += 1

        obs = self._get_obs()
        done = False
        info = {
            'is_success': self._check_success(obs)
        }
        # if info['is_success'] > 0:
        #     print('Success')
        reward = self.compute_reward(obs['achieved_goal'], obs['desired_goal'], info)

        return obs, reward, done, info

    def reset(self):
        super().reset()

        self._destroy_obi_softbody()
        self.env.reset()
        if not self.open_gripper:
            self.articulation_channel.set_action(
                'SetJointPositionDirectly',
                index=1,
                joint_positions=[0, 0],
            )

        obi_softbody_solver_position = self._load_obi_softbody()
        self.goal = self._sample_goal()
        self.game_object_channel.set_action(
            'SetTransform',
            index=0,
            position=list(self.goal * self.scale)
        )

        self.t = 0
        self.ik_controller.reset()
        joint_positions = self.ik_controller.calculate_ik_recursive(
            list(obi_softbody_solver_position),
        )
        self.articulation_channel.set_action(
            'SetJointPositionDirectly',
            index=0,
            joint_positions=list(joint_positions),
        )
        self._step()

        return self._get_obs()

    def seed(self, seed=1234):
        self.np_random, seed = seeding.np_random(seed)
        return [seed]

    def render(self, mode='human'):
        self._step()

    def compute_reward(self, achieved_goal, desired_goal, info):
        distance = self._compute_goal_distance(achieved_goal, desired_goal)
        if self.reward_type == 'sparse':
            return -(distance > self.tolerance).astype(np.float32)
        else:
            return -distance

    def _get_obs(self): #
        gripper_position = np.array(self.articulation_channel.data[1]['positions'][3])
        gripper_velocity = np.array(self.articulation_channel.data[1]['velocities'][3])
        gripper_width = self._get_gripper_width()
        panda_obs = np.concatenate((gripper_position, gripper_velocity, [gripper_width])) / self.scale

        softbody_position = np.array(self.obi_softbody_channel.data[0]['position']) / self.scale
        softbody_orientation = np.array(self.obi_softbody_channel.data[0]['orientation'])
        softbody_velocity = np.array(self.obi_softbody_channel.data[0]['velocity']) / self.scale
        softbody_angular_vel = np.array(self.obi_softbody_channel.data[0]['angular_vel'])
        softbody_obs = np.concatenate(
            (softbody_position, softbody_orientation, softbody_velocity, softbody_angular_vel)
        )

        obs = np.concatenate((panda_obs, softbody_obs))

        return {
            'observation': obs.copy(),
            'achieved_goal': softbody_position.copy(),
            'desired_goal': self.goal.copy()
        }

    def _load_obi_softbody(self):
        if self.load_softbody:
            return
        obi_softbody_solver_position = self.np_random.uniform(
            self.softbody_range_low, self.softbody_range_high
        )
        self.asset_channel.set_action(
            'LoadObiSoftbody',
            filename=self.asset_bundle_file,
            name='Obi Softbody Solver',
            position=list(obi_softbody_solver_position * self.scale)
        )
        self._step()
        self.load_softbody = True

        return obi_softbody_solver_position

    def _destroy_obi_softbody(self):
        if not self.load_softbody:
            return
        self.obi_softbody_channel.set_action(
            'Destroy',
            index=0
        )
        self._step()
        self.load_softbody = False

    def _sample_goal(self):
        goal = self.np_random.uniform(self.goal_range_low, self.goal_range_high)
        return goal.copy()

    def _get_gripper_position(self):
        return np.array(self.articulation_channel.data[1]['positions'][3]) / self.scale

    def _get_gripper_width(self):
        gripper_joint_positions = copy.deepcopy(self.articulation_channel.data[1]['joint_positions'])
        return -1 * (gripper_joint_positions[0] + gripper_joint_positions[1]) / self.scale

    def _set_franka_joints(self, a: np.ndarray):
        self.articulation_channel.set_action(
            'SetJointPosition',
            index=0,
            joint_positions=list(a[0:7]),
        )
        self._step()

        a[7] = -1 * a[7] / 2 * self.scale
        self.articulation_channel.set_action(
            'SetJointPosition',
            index=1,
            joint_positions=[a[7], a[7]],
        )
        self._step()

    def _compute_goal_distance(self, goal_a, goal_b):
        assert goal_a.shape == goal_b.shape
        return np.linalg.norm(goal_a - goal_b, axis=-1)

    def _check_success(self, obs):
        achieved_goal = obs['achieved_goal']
        desired_goal = obs['desired_goal']
        distance = self._compute_goal_distance(achieved_goal, desired_goal)

        return (distance < self.tolerance).astype(np.float32)



from typing import List
import struct


class OutgoingMessage:
    """
    Utility class for forming the message that is written to a SideChannel.
    All data is written in little-endian format using the struct module.
    """

    def __init__(self):
        """
        Create an OutgoingMessage with an empty buffer.
        """
        self.buffer = bytearray()

    def write_bool(self, b: bool) -> None:
        """
        Append a boolean value.
        """
        self.buffer.extend(int(b).to_bytes(1, byteorder='little'))

    def write_int32(self, i: int) -> None:
        """
        Append an integer value.
        """
        self.buffer.extend(i.to_bytes(4, byteorder='little'))

    def write_float32(self, f: float) -> None:
        """
        Append a float value. It will be truncated to 32-bit precision.
        """
        self.buffer.extend(struct.pack('f', f))

    def write_float32_list(self, float_list: List[float]) -> None:
        """
        Append a list of float values. They will be truncated to 32-bit precision.
        """
        self.write_int32(len(float_list))
        for f in float_list:
            self.write_float32(f)

    def write_string(self, s: str) -> None:
        """
        Append a string value. Internally, it will be encoded to ascii, and the
        encoded length will also be written to the message.
        """
        s_byte = s.encode('utf-8')
        self.write_int32(len(s_byte))
        self.buffer.extend(s_byte)



import pyrfuniverse.attributes as attr


class ClothAttr(attr.BaseAttr):
    """
    Obi cloth class.
    """
    def parse_message(self, data: dict):
        """
        Parse messages. This function is called by internal function.

        Returns:
            Dict: A dict containing useful information of this class.

        """
        super().parse_message(data)


import pyrfuniverse.attributes as attr


class ColliderAttr(attr.GameObjectAttr):
    """
    Collider class for objects who have collider in Unity.
    """
    def parse_message(self, data: dict):
        """
        Parse messages. This function is called by internal function.

        Returns:
            Dict: A dict containing useful information of this class.
        """
        super().parse_message(data)

    def GenerateVHACDColider(self):
        """
        Generate convex colliders using VHACD algorithm.
        """
        self._send_data('GenerateVHACDColider')


import os
os.environ["OPENCV_IO_ENABLE_OPENEXR"] = "1"
from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
import pyrfuniverse.utils.depth_processor as dp
import numpy as np
try:
    import open3d as o3d
except ImportError:
    raise Exception('This feature requires open3d, please install with `pip install open3d`')

env = RFUniverseBaseEnv(scene_file='PointCloud.json')

nd_intrinsic_matrix = np.array([[960, 0, 960],
                            [0, 960, 540],
                            [0, 0, 1]])
camera1 = env.GetAttr(698548)
camera1.GetDepthEXR(intrinsic_matrix=nd_intrinsic_matrix)
camera1.GetRGB(intrinsic_matrix=nd_intrinsic_matrix)
camera1.GetID(intrinsic_matrix=nd_intrinsic_matrix)
env.step()

image_rgb = camera1.data['rgb']
image_depth_exr = camera1.data['depth_exr']
local_to_world_matrix = camera1.data['local_to_world_matrix']
point1 = dp.image_bytes_to_point_cloud_intrinsic_matrix(image_rgb, image_depth_exr, nd_intrinsic_matrix, local_to_world_matrix)

camera2 = env.GetAttr(698550)
camera2.GetDepthEXR(intrinsic_matrix=nd_intrinsic_matrix)
camera2.GetRGB(intrinsic_matrix=nd_intrinsic_matrix)
camera2.GetID(intrinsic_matrix=nd_intrinsic_matrix)
env.step()

image_rgb = camera2.data['rgb']
image_depth_exr = camera2.data['depth_exr']
local_to_world_matrix = camera2.data['local_to_world_matrix']
point2 = dp.image_bytes_to_point_cloud_intrinsic_matrix(image_rgb, image_depth_exr, nd_intrinsic_matrix, local_to_world_matrix)
env.close()

# unity space to open3d space and show
point1.transform([[-1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
point2.transform([[-1, 0, 0, 0], [0, 1, 0, 0], [0, 0, 1, 0], [0, 0, 0, 1]])
coorninate = o3d.geometry.TriangleMesh.create_coordinate_frame()
o3d.visualization.draw_geometries([point1, point2, coorninate])




from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
import os

env = RFUniverseBaseEnv()

ur5 = env.LoadURDF(path=os.path.abspath('../URDF/01b24b02-0e4e-11ed-81d4-ec2e98c7e246/motion_unity.urdf'), native_ik=False)
ur5.SetTransform(position=[0, 1, 0])

env.Pend()



import pybullet as p
import pybullet_data
import numpy as np
import math


class RFUniverseStretchController:
    """
    RFUniverseController is a class to generate robot arm joint states. In simulation environment, we mostly
    want to specify the 6DoF of a joint, then the robot arm will automatically move to that state. Thus, here
    we use pybullet.calculateInverseKinematics() to generate joint positions based on a given robot arm, a
    given end-effector joint and a target Cartesian position. The generated joint states will be passed to
    Unity by rfuniverse channels. Besides, this class will also provide functions to align coordinate in Unity
    and in pybullet.
    """

    def __init__(
            self, robot_urdf, base_pos=np.array([0, 0, 0]),
            base_orn=[-0.707107, 0.0, 0.0, 0.707107], init_joint_positions=[0] * 12,
            render=False
    ):
        if render:
            p.connect(p.GUI) # For debug mode
        else:
            p.connect(p.DIRECT)

        p.configureDebugVisualizer(p.COV_ENABLE_Y_AXIS_UP, 1)
        p.setAdditionalSearchPath(pybullet_data.getDataPath())
        p.setGravity(0, -9.8, 0)

        self.bullet_client = p
        self.bullet_flags = self.bullet_client.URDF_ENABLE_CACHED_GRAPHICS_SHAPES

        self.robot_name = 'stretch'
        self.robot_urdf = robot_urdf
        self.end_effector_id = 15
        self.num_dof = 12
        self.init_joint_positions = self.get_pybullet_joint_pos_from_unity(init_joint_positions)
        self.robot = self.bullet_client.loadURDF(self.robot_urdf, base_pos, base_orn, useFixedBase=True,
                                                 flags=self.bullet_flags)

        self.moveable_joint_ids = []
        joint_position_idx = 0
        self.revolute_idx = []
        self.prismatic_idx = []
        for j in range(36):
            self.bullet_client.changeDynamics(self.robot, j, linearDamping=0, angularDamping=0)
            info = self.bullet_client.getJointInfo(self.robot, j)

            jointName = info[1]
            jointType = info[2]
            if jointType == self.bullet_client.JOINT_REVOLUTE or jointType == self.bullet_client.JOINT_PRISMATIC:
                # print(j, jointName)
                self.moveable_joint_ids.append(j)
                if jointType == self.bullet_client.JOINT_REVOLUTE:
                    self.revolute_idx.append(joint_position_idx)
                elif jointType == self.bullet_client.JOINT_PRISMATIC:
                    self.prismatic_idx.append(joint_position_idx)
                joint_position_idx += 1

        self.reset()

        # print('pybullet', self.get_link_state(self.end_effector_id))

    def get_bullet_pos_from_unity(self, unity_pos: list) -> list:
        return [-1 * unity_pos[0], unity_pos[1], unity_pos[2]]

    def get_unity_pos_from_bullet(self, bullet_pos: list) -> list:
        return [-1 * bullet_pos[0], bullet_pos[1], bullet_pos[2]]

    def get_unity_joint_pos_from_pybullet(self, pybullet_joint_pos: tuple) -> list:
        pybullet_joint_pos = list(pybullet_joint_pos)[:self.num_dof]
        for i, (joint_pos) in enumerate(pybullet_joint_pos):
            if i in self.revolute_idx:
                pybullet_joint_pos[i] = 180 * joint_pos / math.pi

        # The order of joints in Unity and Pybullet is not same.
        # So we need to change the order manually here as well as ignoring the rotation of wheel.
        unity_joint_pos = list(pybullet_joint_pos[10:12]) + list(pybullet_joint_pos[2:8]) + \
                          [pybullet_joint_pos[9], pybullet_joint_pos[8]]

        return unity_joint_pos

    def get_pybullet_joint_pos_from_unity(self, unity_joint_pos: list) -> list:
        unity_joint_pos = list(unity_joint_pos)[:self.num_dof]
        unity_joint_pos = np.array(unity_joint_pos)
        pybullet_joint_pos = unity_joint_pos * math.pi / 180

        return pybullet_joint_pos

    def calculate_ik(self, unity_eef_pos, eef_orn=None) -> list:
        if eef_orn is None:
            eef_orn = self.bullet_client.getQuaternionFromEuler([math.pi / 2., 0., 0.])

        eef_pos = self.get_bullet_pos_from_unity(unity_eef_pos)

        joint_positions = self.bullet_client.calculateInverseKinematics(
            self.robot, self.end_effector_id, eef_pos, eef_orn, maxNumIterations=20)

        for i, (idx) in enumerate(self.moveable_joint_ids):
            self.bullet_client.resetJointState(self.robot, idx, joint_positions[i])

        return self.get_unity_joint_pos_from_pybullet(joint_positions)

    def calculate_ik_recursive(self, unity_eef_pos, eef_orn=None) -> list:
        if eef_orn is None:
            eef_orn = self.bullet_client.getQuaternionFromEuler([math.pi / 2., 0., 0.])

        eef_pos = self.get_bullet_pos_from_unity(unity_eef_pos)
        for i in range(20):
            joint_positions = self.bullet_client.calculateInverseKinematics(
                self.robot, self.end_effector_id, eef_pos, eef_orn, maxNumIterations=20)

            for i, (idx) in enumerate(self.moveable_joint_ids):
                self.bullet_client.resetJointState(self.robot, idx, joint_positions[i])

        return self.get_unity_joint_pos_from_pybullet(joint_positions)

    def get_link_state(self, link_idx):
        link_state = self.bullet_client.getLinkState(self.robot, link_idx)

        return self.get_unity_pos_from_bullet(link_state[0])

    def reset(self):
        for i, (idx) in enumerate(self.moveable_joint_ids):
            self.bullet_client.resetJointState(self.robot, idx, self.init_joint_positions[i])



from pyrfuniverse.envs.base_env import RFUniverseBaseEnv
import pyrfuniverse.attributes as attr

env = RFUniverseBaseEnv(assets=['Rigidbody_Box', 'franka_panda'])

box = env.InstanceObject(name='Rigidbody_Box', id=123456, attr_type=attr.RigidbodyAttr)
box.SetTransform(position=[0, 1, 0])
env.step(5)
print('Rigidbody_Box:')
print('position:')
print(box.data['position'])
print('rotation:')
print(box.data['rotation'])
print('quaternion:')
print(box.data['quaternion'])
print('local_position:')
print(box.data['local_position'])
print('local_rotation:')
print(box.data['local_rotation'])
print('local_quaternion:')
print(box.data['local_quaternion'])
print('velocity:')
print(box.data['velocity'])
print('angular_vel:')
print(box.data['angular_vel'])

robot = env.InstanceObject(name='franka_panda', id=789789, attr_type=attr.ControllerAttr)
robot.SetTransform(position=[1, 0, 0])
env.step()
print('franka_panda:')
print('number_of_joints:')
print(robot.data['number_of_joints'])
print('positions:')
print(robot.data['positions'])
print('rotations:')
print(robot.data['rotations'])
print('quaternion:')
print(robot.data['quaternion'])
print('local_positions:')
print(robot.data['local_positions'])
print('local_rotations:')
print(robot.data['local_rotations'])
print('local_quaternion:')
print(robot.data['local_quaternion'])
print('velocities:')
print(robot.data['velocities'])
print('number_of_moveable_joints:')
print(robot.data['number_of_moveable_joints'])
print('joint_positions:')
print(robot.data['joint_positions'])
print('joint_velocities:')
print(robot.data['joint_velocities'])
print('joint_accelerations:')
print(robot.data['joint_accelerations'])
print('joint_force:')
print(robot.data['joint_force'])
print('joint_types:')
print(robot.data['joint_types'])
print('joint_lower_limit:')
print(robot.data['joint_lower_limit'])
print('joint_upper_limit:')
print(robot.data['joint_upper_limit'])
env.close()



import numpy as np
import pyrfuniverse.attributes as attr


class PointCloudAttr(attr.BaseAttr):
    """
    Point cloud rendering class.
    """
    def parse_message(self, data: dict):
        """
        Parse messages. This function is called by internal function.

        Returns:
            Dict: A dict containing useful information of this class.
        """
        super().parse_message(data)

    def ShowPointCloud(self, positions: np.ndarray = None, colors: np.ndarray = None, ply_path: str = None,
                       radius: float = 0.01):
        """
        Display point cloud in Unity.

        Args:
            positions: A list of positions of points in a point cloud.
            colors: A list of colors of points (range [0, 1]) in a point cloud.
            ply_path: Str, the absolute path of `.ply` file. If this parameter is specified, `positions`
                and `colors` will be ignored.
            radius: Float, the radius of the point cloud.
        """
        self._send_data('ShowPointCloud', ply_path, positions, colors, radius)

    def SetRadius(self, radius: float):
        """
        Set the radius for points in a point cloud.

        Args:
            radius: Float, the radius.
        """
        self._send_data('SetRadius', radius)
